# Originally imported via:
#   stubgen {...} -m mlir._mlir_libs._mlirExecutionEngine
# Local modifications:
#   * Relative imports for cross-module references.
#   * Add __all__

from collections.abc import Sequence

from ._mlir import ir as _ir

__all__ = [
    "ExecutionEngine",
]

class ExecutionEngine:
    def __init__(self, module: _ir.Module, opt_level: int = 2, shared_libs: Sequence[str] = ...) -> None: ...
    def _CAPICreate(self) -> object: ...
    def _testing_release(self) -> None: ...
    def dump_to_object_file(self, file_name: str) -> None: ...
    def raw_lookup(self, func_name: str) -> int: ...
    def raw_register_runtime(self, name: str, callback: object) -> None: ...
    def init() -> None: ...
    @property
    def _CAPIPtr(self) -> object: ...



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

#  This file contains functions to convert between Memrefs and NumPy arrays and vice-versa.

import numpy as np
import ctypes

try:
    import ml_dtypes
except ModuleNotFoundError:
    # The third-party ml_dtypes provides some optional low precision data-types for NumPy.
    ml_dtypes = None


class C128(ctypes.Structure):
    """A ctype representation for MLIR's Double Complex."""

    _fields_ = [("real", ctypes.c_double), ("imag", ctypes.c_double)]


class C64(ctypes.Structure):
    """A ctype representation for MLIR's Float Complex."""

    _fields_ = [("real", ctypes.c_float), ("imag", ctypes.c_float)]


class F16(ctypes.Structure):
    """A ctype representation for MLIR's Float16."""

    _fields_ = [("f16", ctypes.c_int16)]


class BF16(ctypes.Structure):
    """A ctype representation for MLIR's BFloat16."""

    _fields_ = [("bf16", ctypes.c_int16)]


class F8E5M2(ctypes.Structure):
    """A ctype representation for MLIR's Float8E5M2."""

    _fields_ = [("f8E5M2", ctypes.c_int8)]


class F8E3M4(ctypes.Structure):
    """A ctype representation for MLIR's Float8E3M4."""

    _fields_ = [("f8E3M4", ctypes.c_int8)]


class F8E4M3(ctypes.Structure):
    """A ctype representation for MLIR's Float8E4M3."""

    _fields_ = [("f8E4M3", ctypes.c_int8)]


# https://stackoverflow.com/questions/26921836/correct-way-to-test-for-numpy-dtype
def as_ctype(dtp):
    """Converts dtype to ctype."""
    if dtp == np.dtype(np.complex128):
        return C128
    if dtp == np.dtype(np.complex64):
        return C64
    if dtp == np.dtype(np.float16):
        return F16
    if ml_dtypes is not None and dtp == ml_dtypes.bfloat16:
        return BF16
    if ml_dtypes is not None and dtp == ml_dtypes.float8_e5m2:
        return F8E5M2
    if ml_dtypes is not None and dtp == ml_dtypes.float8_e3m4:
        return F8E3M4
    if ml_dtypes is not None and dtp == ml_dtypes.float8_e4m3:
        return F8E4M3
    return np.ctypeslib.as_ctypes_type(dtp)


def to_numpy(array):
    """Converts ctypes array back to numpy dtype array."""
    if array.dtype == C128:
        return array.view("complex128")
    if array.dtype == C64:
        return array.view("complex64")
    if array.dtype == F16:
        return array.view("float16")
    assert not (
        array.dtype in (BF16, F8E5M2, F8E3M4, F8E4M3) and ml_dtypes is None
    ), f"{array.dtype=} requires the ml_dtypes package, please run:\n\npip install ml_dtypes\n"
    if array.dtype == BF16:
        return array.view("bfloat16")
    if array.dtype == F8E5M2:
        return array.view("float8_e5m2")
    if array.dtype == F8E3M4:
        return array.view("float8_e3m4")
    if array.dtype == F8E4M3:
        return array.view("float8_e4m3")

    return array


def make_nd_memref_descriptor(rank, dtype):
    class MemRefDescriptor(ctypes.Structure):
        """Builds an empty descriptor for the given rank/dtype, where rank>0."""

        _fields_ = [
            ("allocated", ctypes.c_longlong),
            ("aligned", ctypes.POINTER(dtype)),
            ("offset", ctypes.c_longlong),
            ("shape", ctypes.c_longlong * rank),
            ("strides", ctypes.c_longlong * rank),
        ]

    return MemRefDescriptor


def make_zero_d_memref_descriptor(dtype):
    class MemRefDescriptor(ctypes.Structure):
        """Builds an empty descriptor for the given dtype, where rank=0."""

        _fields_ = [
            ("allocated", ctypes.c_longlong),
            ("aligned", ctypes.POINTER(dtype)),
            ("offset", ctypes.c_longlong),
        ]

    return MemRefDescriptor


class UnrankedMemRefDescriptor(ctypes.Structure):
    """Creates a ctype struct for memref descriptor"""

    _fields_ = [("rank", ctypes.c_longlong), ("descriptor", ctypes.c_void_p)]


def get_ranked_memref_descriptor(nparray):
    """Returns a ranked memref descriptor for the given numpy array."""
    ctp = as_ctype(nparray.dtype)
    if nparray.ndim == 0:
        x = make_zero_d_memref_descriptor(ctp)()
        x.allocated = nparray.ctypes.data
        x.aligned = nparray.ctypes.data_as(ctypes.POINTER(ctp))
        x.offset = ctypes.c_longlong(0)
        return x

    x = make_nd_memref_descriptor(nparray.ndim, ctp)()
    x.allocated = nparray.ctypes.data
    x.aligned = nparray.ctypes.data_as(ctypes.POINTER(ctp))
    x.offset = ctypes.c_longlong(0)
    x.shape = nparray.ctypes.shape

    # Numpy uses byte quantities to express strides, MLIR OTOH uses the
    # torch abstraction which specifies strides in terms of elements.
    strides_ctype_t = ctypes.c_longlong * nparray.ndim
    x.strides = strides_ctype_t(*[x // nparray.itemsize for x in nparray.strides])
    return x


def get_unranked_memref_descriptor(nparray):
    """Returns a generic/unranked memref descriptor for the given numpy array."""
    d = UnrankedMemRefDescriptor()
    d.rank = nparray.ndim
    x = get_ranked_memref_descriptor(nparray)
    d.descriptor = ctypes.cast(ctypes.pointer(x), ctypes.c_void_p)
    return d


def move_aligned_ptr_by_offset(aligned_ptr, offset):
    """Moves the supplied ctypes pointer ahead by `offset` elements."""
    aligned_addr = ctypes.addressof(aligned_ptr.contents)
    elem_size = ctypes.sizeof(aligned_ptr.contents)
    shift = offset * elem_size
    content_ptr = ctypes.cast(aligned_addr + shift, type(aligned_ptr))
    return content_ptr


def unranked_memref_to_numpy(unranked_memref, np_dtype):
    """Converts unranked memrefs to numpy arrays."""
    ctp = as_ctype(np_dtype)
    descriptor = make_nd_memref_descriptor(unranked_memref[0].rank, ctp)
    val = ctypes.cast(unranked_memref[0].descriptor, ctypes.POINTER(descriptor))
    content_ptr = move_aligned_ptr_by_offset(val[0].aligned, val[0].offset)
    np_arr = np.ctypeslib.as_array(content_ptr, shape=val[0].shape)
    strided_arr = np.lib.stride_tricks.as_strided(
        np_arr,
        np.ctypeslib.as_array(val[0].shape),
        np.ctypeslib.as_array(val[0].strides) * np_arr.itemsize,
    )
    return to_numpy(strided_arr)


def ranked_memref_to_numpy(ranked_memref):
    """Converts ranked memrefs to numpy arrays."""
    content_ptr = move_aligned_ptr_by_offset(
        ranked_memref[0].aligned, ranked_memref[0].offset
    )
    np_arr = np.ctypeslib.as_array(content_ptr, shape=ranked_memref[0].shape)
    strided_arr = np.lib.stride_tricks.as_strided(
        np_arr,
        np.ctypeslib.as_array(ranked_memref[0].shape),
        np.ctypeslib.as_array(ranked_memref[0].strides) * np_arr.itemsize,
    )
    return to_numpy(strided_arr)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import functools
from types import MappingProxyType

from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.datamodel.registry import register
from numba_cuda_mlir.numba_cuda.typing.asnumbatype import as_numba_type
from numba_cuda_mlir.numba_cuda.typing.typeof import typeof_impl
from numba_cuda_mlir.numba_cuda.typing.templates import (
    Registry,
    _OverloadAttributeTemplate,
    _OverloadFunctionTemplate,
    _OverloadMethodTemplate,
    make_overload_template,
    make_overload_attribute_template,
)

from numba_cuda_mlir.numba_cuda.extending import (
    intrinsic,
    _Intrinsic,
)

from numba_cuda_mlir.models import mlir_data_manager

from numba_cuda_mlir.lowering_registry import LoweringRegistry
from numba_cuda_mlir.extending.argument_handler import ArgumentHandler
from numba_cuda_mlir._whole_function_planners import (
    WholeFunctionPlanner,
    register_planner,
    require_launch_config,
    set_required_dynamic_shared_memory,
)

lowering_registry = LoweringRegistry()
typing_registry = Registry()
lower_cast = lowering_registry.lower_cast
lower_builtin = lowering_registry.lower
register_model = functools.partial(register, mlir_data_manager)

__all__ = [
    "ArgumentHandler",
    "WholeFunctionPlanner",
    "intrinsic",
    "lowering_registry",
    "as_numba_type",
    "lower_builtin",
    "lower_cast",
    "refresh_registries",
    "register_planner",
    "require_launch_config",
    "set_required_dynamic_shared_memory",
    "overload",
    "overload_attribute",
    "overload_method",
    "register_jitable",
    "type_callable",
    "typing_registry",
    "typeof_impl",
]

_overload_default_jit_options = {"no_cpython_wrapper": True, "nopython": True}


def _require_typing_registry(decorator_name, registry):
    if registry is None:
        raise ValueError(f"numba_cuda_mlir.extending.{decorator_name} requires typing_registry=")
    return registry


def _require_lowering_registry(decorator_name, registry):
    if registry is None:
        raise ValueError(f"numba_cuda_mlir.extending.{decorator_name} requires lowering_registry=")
    return registry


def refresh_registries(*, typing=True, target=True, include_uninitialized_cuda=True):
    """Sync compiler contexts after adding entries to extension registries."""
    from numba_cuda_mlir.descriptor import mlir_target
    from numba_cuda_mlir.numba_cuda.descriptor import cuda_target

    mlir_target.refresh_registries(typing=typing, target=target)
    if (
        include_uninitialized_cuda
        or cuda_target._typingctx_initialized
        or cuda_target._targetctx_initialized
    ):
        cuda_target.refresh_registries(typing=typing, target=target)


def type_callable(func, typing_registry=None):
    """
    Decorate a function as implementing typing for the callable *func*.
    """
    from numba_cuda_mlir.numba_cuda.typing.templates import CallableTemplate

    selected_typing_registry = typing_registry or globals()["typing_registry"]

    if not callable(func) and not isinstance(func, str):
        raise TypeError("`func` should be a function or string")
    try:
        func_name = func.__name__
    except AttributeError:
        func_name = str(func)

    def decorate(typing_func):
        def generic(self):
            return typing_func(self.context)

        name = "%s_CallableTemplate" % (func_name,)
        bases = (CallableTemplate,)
        class_dict = dict(key=func, generic=generic)
        template = type(name, bases, class_dict)
        selected_typing_registry.register(template)
        if callable(func):
            selected_typing_registry.register_global(func, types.Function(template))
        refresh_registries(typing=True, target=False, include_uninitialized_cuda=False)
        return typing_func

    return decorate


class _NumbaCudaMlirOverloadFunctionTemplate(_OverloadFunctionTemplate):
    def _get_jit_decorator(self):
        from numba_cuda_mlir import cuda
        from numba_cuda_mlir.mlir_compiler import get_compiler_class

        def jit_with_mlir_pipeline(**jit_options):
            jit_decorator = cuda.jit(**jit_options)

            def decorate(pyfunc):
                disp = jit_decorator(pyfunc)
                fcomp = getattr(disp, "_compiler", None)
                if fcomp is not None and getattr(fcomp, "pipeline_class", None) is None:
                    fcomp.pipeline_class = get_compiler_class(disp.targetoptions)
                return disp

            return decorate

        return jit_with_mlir_pipeline


def overload(
    func,
    jit_options=MappingProxyType({}),
    strict=True,
    inline="never",
    prefer_literal=False,
    typing_registry=None,
    **kwargs,
):
    """Register an implementation for ``func``.

    The decorated function is a *typer*: it is called at compile time with the
    Numba types of the arguments and must return a Python function (the
    implementation), or ``None`` to decline the overload. The implementation
    is compiled by Numba-CUDA-MLIR's MLIR pipeline.

    Parameters
    ----------
    func
        The Python callable being overloaded.
    jit_options : Mapping
        Options forwarded to ``cuda.jit`` when compiling the implementation.
    strict : bool
        If ``True``, raise when the implementation cannot be compiled. If
        ``False``, the failure is silenced (used by :func:`register_jitable`).
    inline : str
        Inlining policy: ``"never"``, ``"always"``, or a cost-model callable.
    prefer_literal : bool
        If ``True``, prefer literal-typed arguments when resolving the
        overload.
    """
    selected_typing_registry = _require_typing_registry("overload", typing_registry)
    jit_options = dict(jit_options)
    opts = _overload_default_jit_options.copy()
    opts.update(jit_options)

    def decorate(overload_func):
        template = make_overload_template(
            func,
            overload_func,
            opts,
            strict,
            inline,
            prefer_literal,
            base=_NumbaCudaMlirOverloadFunctionTemplate,
            **kwargs,
        )
        template._typing_registry = selected_typing_registry
        selected_typing_registry.register(template)
        if callable(func):
            selected_typing_registry.register_global(func, types.Function(template))
        refresh_registries(typing=True, target=False, include_uninitialized_cuda=False)
        return overload_func

    return decorate


class _NumbaCudaMlirOverloadAttributeTemplate(_OverloadAttributeTemplate):
    """Override _init_once to register a numba_cuda_mlir-compatible lower_getattr."""

    _lowering_registered = set()

    def _init_once(self):
        cls = type(self)
        if cls in _NumbaCudaMlirOverloadAttributeTemplate._lowering_registered:
            return
        _NumbaCudaMlirOverloadAttributeTemplate._lowering_registered.add(cls)

        attr = cls._attr
        key = cls.key
        selected_lowering_registry = cls._attribute_lowering_registry

        @selected_lowering_registry.lower_getattr(key, attr)
        def getattr_impl(context, builder, target, value):
            value_type = builder.get_numba_type(value.name)
            disp = cls._find_overload_dispatcher(context.typing_context, value_type)
            if disp is None:
                raise NotImplementedError(
                    f"No overload_attribute dispatcher for {value_type}.{attr}"
                )
            builder.lower_overload_call(target, disp, [value])

    @classmethod
    def _find_overload_dispatcher(cls, typing_context, typ):
        """Find the cached overload Dispatcher for the given type."""
        overload_func = cls._overload_func
        fnty = typing_context.resolve_value_type(overload_func)
        for temp_cls in getattr(fnty, "templates", []):
            if not hasattr(temp_cls, "_impl_cache"):
                continue
            for cache_key, cache_value in temp_cls._impl_cache.items():
                if cache_value is None or len(cache_key) != 4:
                    continue
                _, args, _, _ = cache_key
                if args == (typ,):
                    disp, _ = cache_value
                    if hasattr(disp, "py_func"):
                        return disp
        return None


class _NumbaCudaMlirOverloadMethodTemplate(_OverloadMethodTemplate):
    """Override _init_once to skip numba's llvmlite lowering registration.

    Method lowering in numba_cuda_mlir goes through BoundFunction → get_overload_builder,
    so the registry-based lowering from numba's _init_once is never used.
    Skipping it avoids polluting numba_cuda_mlir's lowering registry with numba closures.
    """

    def _init_once(self):
        pass


def overload_attribute(typ, attr, typing_registry=None, lowering_registry=None, **kwargs):
    """Register an implementation for a read-only attribute on a Numba type.

    The decorated function is a typer with the same contract as
    :func:`overload`. Its only parameter is the receiver, and the returned
    implementation is a function of the receiver that produces the attribute
    value.

    Parameters
    ----------
    typ
        The Numba type on which the attribute is being defined.
    attr : str
        The name of the attribute being defined.
    """

    selected_typing_registry = _require_typing_registry("overload_attribute", typing_registry)
    selected_lowering_registry = _require_lowering_registry("overload_attribute", lowering_registry)

    def decorate(overload_func):
        template = make_overload_attribute_template(
            typ,
            attr,
            overload_func,
            base=_NumbaCudaMlirOverloadAttributeTemplate,
            **kwargs,
        )
        template._attribute_lowering_registry = selected_lowering_registry
        selected_typing_registry.register_attr(template)
        overload(overload_func, typing_registry=selected_typing_registry, **kwargs)(overload_func)
        refresh_registries(typing=True, target=True, include_uninitialized_cuda=False)
        return overload_func

    return decorate


def overload_method(typ, attr, typing_registry=None, **kwargs):
    """Register an implementation for a method on a Numba type.

    The decorated function is a typer with the same contract as
    :func:`overload`. Its first parameter is the receiver (``self``); any
    additional parameters become method arguments.

    Parameters
    ----------
    typ
        The Numba type on which the method is being defined.
    attr : str
        The name of the method being defined.
    """

    selected_typing_registry = _require_typing_registry("overload_method", typing_registry)

    def decorate(overload_func):
        copied_kwargs = kwargs.copy()
        template = make_overload_attribute_template(
            typ,
            attr,
            overload_func,
            inline=copied_kwargs.pop("inline", "never"),
            prefer_literal=copied_kwargs.pop("prefer_literal", False),
            base=_NumbaCudaMlirOverloadMethodTemplate,
            **copied_kwargs,
        )
        selected_typing_registry.register_attr(template)
        overload(overload_func, typing_registry=selected_typing_registry, **kwargs)(overload_func)
        refresh_registries(typing=True, target=False, include_uninitialized_cuda=False)
        return overload_func

    return decorate


def overload_classmethod(typ, attr, typing_registry=None, **kwargs):
    return overload_method(types.TypeRef(typ), attr, typing_registry=typing_registry, **kwargs)


def register_jitable(*args, typing_registry=None, **kwargs):
    """Mark a plain Python function as compilable from device code.

    The function is registered as a non-strict overload of itself, so calls
    to it from inside a kernel or device function dispatch to the original
    Python source compiled by Numba-CUDA-MLIR. ``@register_jitable`` functions
    may call other ``@register_jitable`` functions, ``@cuda.jit`` device
    functions, and any built-in or overloaded operation the compiler
    understands.

    May be used with or without parentheses::

        @register_jitable
        def f(x): ...


        @register_jitable(inline="always")
        def g(x): ...
    """

    selected_typing_registry = _require_typing_registry("register_jitable", typing_registry)

    def wrap(fn):
        copied_kwargs = kwargs.copy()
        inline = copied_kwargs.pop("inline", "never")

        @overload(
            fn,
            jit_options=copied_kwargs,
            inline=inline,
            strict=False,
            typing_registry=selected_typing_registry,
        )
        def ov_wrap(*args, **kwargs):
            return fn

        fn.__numba_cuda_mlir_jitable__ = True
        return fn

    if kwargs or not args:
        return wrap
    if len(args) != 1:
        raise TypeError("register_jitable accepts at most one positional argument")
    return wrap(*args)


def make_attribute_wrapper(typeclass, struct_attr, python_attr):
    """
    Make an automatic attribute wrapper exposing member named *struct_attr*
    as a read-only attribute named *python_attr*.
    The given *typeclass*'s model must be a StructModel subclass.

    Vendored from cusimt.extending with a change to consider the CUDA data
    model manager.
    """
    from numba_cuda_mlir.numba_cuda.typing.templates import AttributeTemplate
    from numba_cuda_mlir.models import StructModel
    from numba_cuda_mlir.numba_cuda.core.imputils import impl_ret_borrowed

    from numba_cuda_mlir.typing.builtin import registry as cuda_registry
    from numba_cuda_mlir.lowering.builtins import registry as cuda_impl_registry

    data_model_manager = mlir_data_manager

    if not isinstance(typeclass, type) or not issubclass(typeclass, types.Type):
        raise TypeError("typeclass should be a Type subclass, got %s" % (typeclass,))

    def get_attr_fe_type(typ):
        """
        Get the Numba type of member *struct_attr* in *typ*.
        """
        model = data_model_manager.lookup(typ)
        if not isinstance(model, StructModel):
            raise TypeError(
                "make_struct_attribute_wrapper() needs a type "
                "with a StructModel, but got %s" % (model,)
            )
        return model.get_member_fe_type(struct_attr)

    @cuda_registry.register_attr
    class StructAttribute(AttributeTemplate):
        key = typeclass

        def generic_resolve(self, typ, attr):
            if attr == python_attr:
                return get_attr_fe_type(typ)

    @cuda_impl_registry.lower_getattr(typeclass, python_attr)
    def struct_getattr_impl(context, builder, target, value):
        from numba_cuda_mlir._mlir import ir as mlir_ir
        from numba_cuda_mlir._mlir.dialects import llvm as llvm_dialect

        value_type = builder.get_numba_type(value.name)
        model = data_model_manager.lookup(value_type)
        field_idx = model.get_field_position(struct_attr)
        field_mlir_type = model.get_model(field_idx).get_value_type()
        struct_val = builder.load_var(value)
        extracted = llvm_dialect.extractvalue(
            field_mlir_type,
            struct_val,
            position=mlir_ir.DenseI64ArrayAttr.get([field_idx]),
        )
        builder.store_var(target, extracted)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from collections import namedtuple
import math
from functools import reduce

import numpy as np
import operator

from numba_cuda_mlir.numba_cuda.core.imputils import (
    call_getiter,
    call_iternext,
    impl_ret_borrowed,
    impl_ret_untracked,
    numba_typeref_ctor,
    Registry,
)
from numba_cuda_mlir.numba_cuda import typing, types
from numba_cuda_mlir.numba_cuda import cgutils
from numba_cuda_mlir.numba_cuda.extending import overload, intrinsic, register_jitable
from numba_cuda_mlir.numba_cuda.core.errors import (
    TypingError,
    LoweringError,
    NumbaTypeError,
    RequireLiteralValue,
)
from numba_cuda_mlir.numba_cuda.typing.templates import AbstractTemplate, signature

# Several @lower codegen here (const eq/ne, round, get_type_min/max_value) built
# llvmlite IR and are filtered out on the MLIR path; neutered to raise this.
_DEAD_CODEGEN_MSG = "this builtins llvmlite codegen is not used on the MLIR path"
from numba_cuda_mlir.numba_cuda.typing.templates import infer_global
from numba_cuda_mlir.numba_cuda.misc.special import literal_unroll
from numba_cuda_mlir.numba_cuda.typing.asnumbatype import as_numba_type

from numba_cuda_mlir.numba_cuda.typing.builtins import IndexValue, IndexValueType


registry = Registry("cpython.builtins")
lower = registry.lower
lower_cast = registry.lower_cast
lower_constant = registry.lower_constant
lower_getattr_generic = registry.lower_getattr_generic


@overload(operator.truth)
def ol_truth(val):
    if isinstance(val, types.Boolean):

        def impl(val):
            return val

        return impl


@lower(operator.is_not, types.Any, types.Any)
def generic_is_not(context, builder, sig, args):
    """
    Implement `x is not y` as `not (x is y)`.
    """
    is_impl = context.get_function(operator.is_, sig)
    return builder.not_(is_impl(builder, args))


@lower(operator.is_, types.Any, types.Any)
def generic_is(context, builder, sig, args):
    """
    Default implementation for `x is y`
    """
    lhs_type, rhs_type = sig.args
    # the lhs and rhs have the same type
    if lhs_type == rhs_type:
        # mutable types
        if lhs_type.mutable:
            msg = "no default `is` implementation"
            raise LoweringError(msg)
        # immutable types
        else:
            # fallbacks to `==`
            try:
                eq_impl = context.get_function(operator.eq, sig)
            except NotImplementedError:
                # no `==` implemented for this type
                return cgutils.false_bit
            else:
                return eq_impl(builder, args)
    else:
        return cgutils.false_bit


@lower(operator.is_, types.Opaque, types.Opaque)
def opaque_is(context, builder, sig, args):
    """
    Implementation for `x is y` for Opaque types.
    """
    lhs_type, rhs_type = sig.args
    # the lhs and rhs have the same type
    if lhs_type == rhs_type:
        lhs_ptr = builder.ptrtoint(args[0], cgutils.intp_t)
        rhs_ptr = builder.ptrtoint(args[1], cgutils.intp_t)

        return builder.icmp_unsigned("==", lhs_ptr, rhs_ptr)
    else:
        return cgutils.false_bit


@lower(operator.is_, types.Boolean, types.Boolean)
def bool_is_impl(context, builder, sig, args):
    """
    Implementation for `x is y` for types derived from types.Boolean
    (e.g. BooleanLiteral), and cross-checks between literal and non-literal
    booleans, to satisfy Python's behavior preserving identity for bools.
    """
    arg1, arg2 = args
    arg1_type, arg2_type = sig.args
    _arg1 = context.cast(builder, arg1, arg1_type, types.boolean)
    _arg2 = context.cast(builder, arg2, arg2_type, types.boolean)
    eq_impl = context.get_function(
        operator.eq,
        typing.signature(types.boolean, types.boolean, types.boolean),
    )
    return eq_impl(builder, (_arg1, _arg2))


# keep types.IntegerLiteral, as otherwise there's ambiguity between this and int_eq_impl
@lower(operator.eq, types.Literal, types.Literal)
@lower(operator.eq, types.IntegerLiteral, types.IntegerLiteral)
def const_eq_impl(context, builder, sig, args):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


# keep types.IntegerLiteral, as otherwise there's ambiguity between this and int_ne_impl
@lower(operator.ne, types.Literal, types.Literal)
@lower(operator.ne, types.IntegerLiteral, types.IntegerLiteral)
def const_ne_impl(context, builder, sig, args):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


def gen_non_eq(val):
    def none_equality(a, b):
        a_none = isinstance(a, types.NoneType)
        b_none = isinstance(b, types.NoneType)
        if a_none and b_none:

            def impl(a, b):
                return val

            return impl
        elif a_none ^ b_none:

            def impl(a, b):
                return not val

            return impl

    return none_equality


overload(operator.eq)(gen_non_eq(True))
overload(operator.ne)(gen_non_eq(False))

# -------------------------------------------------------------------------------


@lower_getattr_generic(types.DeferredType)
def deferred_getattr(context, builder, typ, value, attr):
    """
    Deferred.__getattr__ => redirect to the actual type.
    """
    inner_type = typ.get()
    val = context.cast(builder, value, typ, inner_type)
    imp = context.get_getattr(inner_type, attr)
    return imp(context, builder, inner_type, val, attr)


@lower_cast(types.Any, types.DeferredType)
@lower_cast(types.Optional, types.DeferredType)
@lower_cast(types.Boolean, types.DeferredType)
def any_to_deferred(context, builder, fromty, toty, val):
    actual = context.cast(builder, val, fromty, toty.get())
    model = context.data_model_manager[toty]
    return model.set(builder, model.make_uninitialized(), actual)


@lower_cast(types.DeferredType, types.Any)
@lower_cast(types.DeferredType, types.Boolean)
@lower_cast(types.DeferredType, types.Optional)
def deferred_to_any(context, builder, fromty, toty, val):
    model = context.data_model_manager[fromty]
    val = model.get(builder, val)
    return context.cast(builder, val, fromty.get(), toty)


# ------------------------------------------------------------------------------


@lower(operator.getitem, types.CPointer, types.Integer)
def getitem_cpointer(context, builder, sig, args):
    base_ptr, idx = args
    elem_ptr = builder.gep(base_ptr, [idx])
    res = builder.load(elem_ptr)
    return impl_ret_borrowed(context, builder, sig.return_type, res)


@lower(operator.setitem, types.CPointer, types.Integer, types.Any)
def setitem_cpointer(context, builder, sig, args):
    base_ptr, idx, val = args
    elem_ptr = builder.gep(base_ptr, [idx])
    builder.store(val, elem_ptr)


# -------------------------------------------------------------------------------


def do_minmax(context, builder, argtys, args, cmpop):
    assert len(argtys) == len(args), (argtys, args)
    assert len(args) > 0

    def binary_minmax(accumulator, value):
        # This is careful to reproduce Python's algorithm, e.g.
        # max(1.5, nan, 2.5) should return 2.5 (not nan or 1.5)
        accty, acc = accumulator
        vty, v = value
        ty = context.typing_context.unify_types(accty, vty)
        assert ty is not None
        acc = context.cast(builder, acc, accty, ty)
        v = context.cast(builder, v, vty, ty)
        cmpsig = typing.signature(types.boolean, ty, ty)
        ge = context.get_function(cmpop, cmpsig)
        pred = ge(builder, (v, acc))
        res = builder.select(pred, v, acc)
        return ty, res

    typvals = zip(argtys, args)
    resty, resval = reduce(binary_minmax, typvals)
    return resval


@lower(max, types.BaseTuple)
def max_iterable(context, builder, sig, args):
    argtys = list(sig.args[0])
    args = cgutils.unpack_tuple(builder, args[0])
    return do_minmax(context, builder, argtys, args, operator.gt)


@lower(max, types.VarArg(types.Any))
def max_vararg(context, builder, sig, args):
    return do_minmax(context, builder, sig.args, args, operator.gt)


@lower(min, types.BaseTuple)
def min_iterable(context, builder, sig, args):
    argtys = list(sig.args[0])
    args = cgutils.unpack_tuple(builder, args[0])
    return do_minmax(context, builder, argtys, args, operator.lt)


@lower(min, types.VarArg(types.Any))
def min_vararg(context, builder, sig, args):
    return do_minmax(context, builder, sig.args, args, operator.lt)


def _round_intrinsic(tp):
    # round() rounds half to even
    return "llvm.rint.f%d" % (tp.bitwidth,)


@lower(round, types.Float)
def round_impl_unary(context, builder, sig, args):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


@lower(round, types.Float, types.Integer)
def round_impl_binary(context, builder, sig, args):
    fltty = sig.args[0]
    # Allow calling the intrinsic from the Python implementation below.
    # This avoids the conversion to an int in Python 3's unary round().
    _round = types.ExternalFunction(_round_intrinsic(fltty), typing.signature(fltty, fltty))

    def round_ndigits(x, ndigits):
        if math.isinf(x) or math.isnan(x):
            return x

        if ndigits >= 0:
            if ndigits > 22:
                # pow1 and pow2 are each safe from overflow, but
                # pow1*pow2 ~= pow(10.0, ndigits) might overflow.
                pow1 = 10.0 ** (ndigits - 22)
                pow2 = 1e22
            else:
                pow1 = 10.0**ndigits
                pow2 = 1.0
            y = (x * pow1) * pow2
            if math.isinf(y):
                return x
            return (_round(y) / pow2) / pow1

        else:
            pow1 = 10.0 ** (-ndigits)
            y = x / pow1
            return _round(y) * pow1

    res = context.compile_internal(builder, round_ndigits, sig, args)
    return impl_ret_untracked(context, builder, sig.return_type, res)


# -------------------------------------------------------------------------------
# Numeric constructors


@lower(int, types.Any)
@lower(float, types.Any)
def int_impl(context, builder, sig, args):
    [ty] = sig.args
    [val] = args
    res = context.cast(builder, val, ty, sig.return_type)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(float, types.StringLiteral)
def float_literal_impl(context, builder, sig, args):
    [ty] = sig.args
    res = context.get_constant(sig.return_type, float(ty.literal_value))
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(complex, types.VarArg(types.Any))
def complex_impl(context, builder, sig, args):
    complex_type = sig.return_type
    float_type = complex_type.underlying_float
    if len(sig.args) == 1:
        [argty] = sig.args
        [arg] = args
        if isinstance(argty, types.Complex):
            # Cast Complex* to Complex*
            res = context.cast(builder, arg, argty, complex_type)
            return impl_ret_untracked(context, builder, sig.return_type, res)
        else:
            real = context.cast(builder, arg, argty, float_type)
            imag = context.get_constant(float_type, 0)

    elif len(sig.args) == 2:
        [realty, imagty] = sig.args
        [real, imag] = args
        real = context.cast(builder, real, realty, float_type)
        imag = context.cast(builder, imag, imagty, float_type)

    cmplx = context.make_complex(builder, complex_type)
    cmplx.real = real
    cmplx.imag = imag
    res = cmplx._getvalue()
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(types.NumberClass, types.Any)
def number_constructor(context, builder, sig, args):
    """
    Call a number class, e.g. np.int32(...)
    """
    if isinstance(sig.return_type, types.Array):
        # Array constructor
        dt = sig.return_type.dtype

        def foo(*arg_hack):
            return np.array(arg_hack, dtype=dt)

        res = context.compile_internal(builder, foo, sig, args)
        return impl_ret_untracked(context, builder, sig.return_type, res)
    else:
        # Scalar constructor
        [val] = args
        [valty] = sig.args
        return context.cast(builder, val, valty, sig.return_type)


# -------------------------------------------------------------------------------
# Constants


@lower_constant(types.Dummy)
def constant_dummy(context, builder, ty, pyval):
    # This handles None, etc.
    return context.get_dummy_value()


@lower_constant(types.ExternalFunctionPointer)
def constant_function_pointer(context, builder, ty, pyval):
    ptrty = context.get_function_pointer_type(ty)
    ptrval = context.add_dynamic_addr(builder, ty.get_pointer(pyval), info=str(pyval))
    return builder.bitcast(ptrval, ptrty)


@lower_constant(types.Optional)
def constant_optional(context, builder, ty, pyval):
    if pyval is None:
        return context.make_optional_none(builder, ty.type)
    else:
        return context.make_optional_value(builder, ty.type, pyval)


# -----------------------------------------------------------------------------


@lower(type, types.Any)
def type_impl(context, builder, sig, args):
    """
    One-argument type() builtin.
    """
    return context.get_dummy_value()


@lower(iter, types.IterableType)
def iter_impl(context, builder, sig, args):
    (ty,) = sig.args
    (val,) = args
    iterval = call_getiter(context, builder, ty, val)
    return iterval


@lower(next, types.IteratorType)
def next_impl(context, builder, sig, args):
    (iterty,) = sig.args
    (iterval,) = args

    res = call_iternext(context, builder, iterty, iterval)

    with builder.if_then(builder.not_(res.is_valid()), likely=False):
        context.fndesc.call_conv.return_user_exc(builder, StopIteration, ())

    return res.yielded_value()


# -----------------------------------------------------------------------------


@lower("not in", types.Any, types.Any)
def not_in(context, builder, sig, args):
    def in_impl(a, b):
        return operator.contains(b, a)

    res = context.compile_internal(builder, in_impl, sig, args)
    return builder.not_(res)


# -----------------------------------------------------------------------------


@lower(len, types.ConstSized)
def constsized_len(context, builder, sig, args):
    [ty] = sig.args
    retty = sig.return_type
    res = context.get_constant(retty, len(ty.types))
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(bool, types.Sized)
def sized_bool(context, builder, sig, args):
    [ty] = sig.args
    if len(ty):
        return cgutils.true_bit
    else:
        return cgutils.false_bit


@lower(tuple)
def lower_empty_tuple(context, builder, sig, args):
    retty = sig.return_type
    res = context.get_constant_undef(retty)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower(tuple, types.BaseTuple)
def lower_tuple(context, builder, sig, args):
    (val,) = args
    return impl_ret_borrowed(context, builder, sig.return_type, val)


@overload(bool)
def bool_sequence(x):
    valid_types = (
        types.CharSeq,
        types.UnicodeCharSeq,
        types.DictType,
        types.ListType,
        types.UnicodeType,
        types.Set,
    )

    if isinstance(x, valid_types):

        def bool_impl(x):
            return len(x) > 0

        return bool_impl


@overload(bool, inline="always")
def bool_none(x):
    if isinstance(x, types.NoneType) or x is None:
        return lambda x: False


# -----------------------------------------------------------------------------


def get_type_max_value(typ):
    if isinstance(typ, types.Float):
        return np.inf
    if isinstance(typ, types.Integer):
        return typ.maxval
    raise NotImplementedError("Unsupported type")


def get_type_min_value(typ):
    if isinstance(typ, types.Float):
        return -np.inf
    if isinstance(typ, types.Integer):
        return typ.minval
    raise NotImplementedError("Unsupported type")


@infer_global(get_type_min_value)
@infer_global(get_type_max_value)
class MinValInfer(AbstractTemplate):
    def generic(self, args, kws):
        assert not kws
        assert len(args) == 1
        if isinstance(args[0], (types.DType, types.NumberClass)):
            return signature(args[0].dtype, *args)


@lower(get_type_min_value, types.NumberClass)
@lower(get_type_min_value, types.DType)
def lower_get_type_min_value(context, builder, sig, args):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


@lower(get_type_max_value, types.NumberClass)
@lower(get_type_max_value, types.DType)
def lower_get_type_max_value(context, builder, sig, args):
    raise NotImplementedError(_DEAD_CODEGEN_MSG)


# -----------------------------------------------------------------------------


@lower(IndexValue, types.intp, types.Type)
@lower(IndexValue, types.uintp, types.Type)
def impl_index_value(context, builder, sig, args):
    typ = sig.return_type
    index, value = args
    index_value = cgutils.create_struct_proxy(typ)(context, builder)
    index_value.index = index
    index_value.value = value
    return index_value._getvalue()


@overload(min)
def indval_min(indval1, indval2):
    if isinstance(indval1, IndexValueType) and isinstance(indval2, IndexValueType):

        def min_impl(indval1, indval2):
            if np.isnan(indval1.value):
                if np.isnan(indval2.value):
                    # both indval1 and indval2 are nans so order by index
                    if indval1.index < indval2.index:
                        return indval1
                    else:
                        return indval2
                else:
                    # comparing against one nan always considered less
                    return indval1
            elif np.isnan(indval2.value):
                # indval1 not a nan but indval2 is so consider indval2 less
                return indval2
            elif indval1.value > indval2.value:
                return indval2
            elif indval1.value == indval2.value:
                if indval1.index < indval2.index:
                    return indval1
                else:
                    return indval2
            return indval1

        return min_impl


@overload(min)
def boolval_min(val1, val2):
    if isinstance(val1, types.Boolean) and isinstance(val2, types.Boolean):

        def bool_min_impl(val1, val2):
            return val1 and val2

        return bool_min_impl


@overload(max)
def indval_max(indval1, indval2):
    if isinstance(indval1, IndexValueType) and isinstance(indval2, IndexValueType):

        def max_impl(indval1, indval2):
            if np.isnan(indval1.value):
                if np.isnan(indval2.value):
                    # both indval1 and indval2 are nans so order by index
                    if indval1.index < indval2.index:
                        return indval1
                    else:
                        return indval2
                else:
                    # comparing against one nan always considered larger
                    return indval1
            elif np.isnan(indval2.value):
                # indval1 not a nan but indval2 is so consider indval2 larger
                return indval2
            elif indval2.value > indval1.value:
                return indval2
            elif indval1.value == indval2.value:
                if indval1.index < indval2.index:
                    return indval1
                else:
                    return indval2
            return indval1

        return max_impl


@overload(max)
def boolval_max(val1, val2):
    if isinstance(val1, types.Boolean) and isinstance(val2, types.Boolean):

        def bool_max_impl(val1, val2):
            return val1 or val2

        return bool_max_impl


greater_than = register_jitable(lambda a, b: a > b)
less_than = register_jitable(lambda a, b: a < b)


@register_jitable
def min_max_impl(iterable, op):
    if isinstance(iterable, types.IterableType):

        def impl(iterable):
            it = iter(iterable)
            return_val = next(it)
            for val in it:
                if op(val, return_val):
                    return_val = val
            return return_val

        return impl


@overload(min)
def iterable_min(iterable):
    return min_max_impl(iterable, less_than)


@overload(max)
def iterable_max(iterable):
    return min_max_impl(iterable, greater_than)


@lower(types.TypeRef, types.VarArg(types.Any))
def redirect_type_ctor(context, builder, sig, args):
    """Redirect constructor implementation to `numba_typeref_ctor(cls, *args)`,
    which should be overloaded by the type's implementation.

    For example:

        d = Dict()

    `d` will be typed as `TypeRef[DictType]()`.  Thus, it will call into this
    implementation.  We need to redirect the lowering to a function
    named ``numba_typeref_ctor``.
    """
    cls = sig.return_type

    def call_ctor(cls, *args):
        return numba_typeref_ctor(cls, *args)

    # Pack arguments into a tuple for `*args`
    ctor_args = types.Tuple.from_types(sig.args)
    # Make signature T(TypeRef[T], *args) where T is cls
    sig = typing.signature(cls, types.TypeRef(cls), ctor_args)
    if len(ctor_args) > 0:
        args = (
            context.get_dummy_value(),  # Type object has no runtime repr.
            context.make_tuple(builder, ctor_args, args),
        )
    else:
        args = (
            context.get_dummy_value(),  # Type object has no runtime repr.
            context.make_tuple(builder, ctor_args, ()),
        )

    return context.compile_internal(builder, call_ctor, sig, args)


@overload(sum)
def ol_sum(iterable, start=0):
    # Cpython explicitly rejects strings, bytes and bytearrays
    # https://github.com/python/cpython/blob/3.9/Python/bltinmodule.c#L2310-L2329 # noqa: E501
    error = None
    if isinstance(start, types.UnicodeType):
        error = ("strings", "")
    elif isinstance(start, types.Bytes):
        error = ("bytes", "b")
    elif isinstance(start, types.ByteArray):
        error = ("bytearray", "b")

    if error is not None:
        msg = "sum() can't sum {} [use {}''.join(seq) instead]".format(*error)
        raise TypingError(msg)

    # if the container is homogeneous then it's relatively easy to handle.
    if isinstance(
        iterable,
        (
            types.containers._HomogeneousTuple,
            types.List,
            types.ListType,
            types.Array,
            types.RangeType,
        ),
    ):
        iterator = iter
    elif isinstance(iterable, (types.containers._HeterogeneousTuple)):
        # if container is heterogeneous then literal unroll and hope for the
        # best.
        iterator = literal_unroll
    else:
        return None

    def impl(iterable, start=0):
        acc = start
        for x in iterator(iterable):
            # This most likely widens the type, this is expected Numba behaviour
            acc = acc + x
        return acc

    return impl


# ------------------------------------------------------------------------------
# map, filter, reduce


@overload(map)
def ol_map(func, iterable, *args):
    def impl(func, iterable, *args):
        for x in zip(iterable, *args):
            yield func(*x)

    return impl


@overload(filter)
def ol_filter(func, iterable):
    if (func is None) or isinstance(func, types.NoneType):

        def impl(func, iterable):
            for x in iterable:
                if x:
                    yield x

    else:

        def impl(func, iterable):
            for x in iterable:
                if func(x):
                    yield x

    return impl


@overload(isinstance)
def ol_isinstance(var, typs):
    def true_impl(var, typs):
        return True

    def false_impl(var, typs):
        return False

    var_ty = as_numba_type(var)

    if isinstance(var_ty, types.Optional):
        msg = f'isinstance cannot handle optional types. Found: "{var_ty}"'
        raise NumbaTypeError(msg)

    # NOTE: The current implementation of `isinstance` restricts the type of the
    # instance variable to types that are well known and in common use. The
    # danger of unrestricted type comparison is that a "default" of `False` is
    # required and this means that if there is a bug in the logic of the
    # comparison tree `isinstance` returns False! It's therefore safer to just
    # reject the compilation as untypable!
    supported_var_ty = (
        types.Number,
        types.Bytes,
        types.RangeType,
        types.DictType,
        types.LiteralStrKeyDict,
        types.List,
        types.ListType,
        types.Tuple,
        types.UniTuple,
        types.Set,
        types.Function,
        types.ClassType,
        types.UnicodeType,
        types.ClassInstanceType,
        types.NoneType,
        types.Array,
        types.Boolean,
        types.Float,
        types.UnicodeCharSeq,
        types.Complex,
        types.NPDatetime,
        types.NPTimedelta,
    )
    if not isinstance(var_ty, supported_var_ty):
        msg = f'isinstance() does not support variables of type "{var_ty}".'
        raise NumbaTypeError(msg)

    t_typs = typs

    # Check the types that the var can be an instance of, it'll be a scalar,
    # a unituple or a tuple.
    if isinstance(t_typs, types.UniTuple):
        # corner case - all types in isinstance are the same
        t_typs = t_typs.key[0]

    if not isinstance(t_typs, types.Tuple):
        t_typs = (t_typs,)

    for typ in t_typs:
        if isinstance(typ, types.Function):
            key = typ.key[0]  # functions like int(..), float(..), str(..)
        elif isinstance(typ, types.ClassType):
            key = typ  # jitclasses
        else:
            key = typ.key

        # corner cases for bytes, range, ...
        # avoid registering those types on `as_numba_type`
        types_not_registered = {
            bytes: types.Bytes,
            range: types.RangeType,
            dict: (types.DictType, types.LiteralStrKeyDict),
            list: types.List,
            tuple: types.BaseTuple,
            set: types.Set,
        }
        if key in types_not_registered:
            if isinstance(var_ty, types_not_registered[key]):
                return true_impl
            continue

        if isinstance(typ, types.TypeRef):
            # Use of Numba type classes is in general not supported as they do
            # not work when the jit is disabled.
            if key not in (types.ListType, types.DictType):
                msg = "Numba type classes (except numba.typed.* container types) are not supported."
                raise NumbaTypeError(msg)
            # Case for TypeRef (i.e. isinstance(var, typed.List))
            #      var_ty == ListType[int64] (instance)
            #         typ == types.ListType  (class)
            return true_impl if type(var_ty) is key else false_impl
        else:
            numba_typ = as_numba_type(key)
            if var_ty == numba_typ:
                return true_impl
            elif isinstance(numba_typ, (types.NPDatetime, types.NPTimedelta)):
                if isinstance(var_ty, type(numba_typ)):
                    return true_impl
            elif (
                isinstance(numba_typ, types.ClassType)
                and isinstance(var_ty, types.ClassInstanceType)
                and var_ty.key == numba_typ.instance_type.key
            ):
                # check for jitclasses
                return true_impl
            elif isinstance(numba_typ, types.Container) and numba_typ.key[0] == types.undefined:
                # check for containers (list, tuple, set, ...)
                if isinstance(var_ty, numba_typ.__class__) or (
                    isinstance(var_ty, types.BaseTuple) and isinstance(numba_typ, types.BaseTuple)
                ):
                    return true_impl

    return false_impl


# -- getattr implementation


def _getattr_raise_attr_exc(obj, name):
    # Dummy function for the purpose of creating an overloadable stub from
    # which to raise an AttributeError as needed
    pass


@overload(_getattr_raise_attr_exc)
def ol__getattr_raise_attr_exc(obj, name):
    if not isinstance(name, types.StringLiteral):
        raise RequireLiteralValue("argument 'name' must be a literal string")
    lname = name.literal_value
    message = f"'{obj}' has no attribute '{lname}'"

    def impl(obj, name):
        raise AttributeError(message)

    return impl


@intrinsic
def resolve_getattr(tyctx, obj, name, default):
    if not isinstance(name, types.StringLiteral):
        raise RequireLiteralValue("argument 'name' must be a literal string")
    lname = name.literal_value
    fn = tyctx.resolve_getattr(obj, lname)
    # Cannot handle things like `getattr(np, 'cos')` as the return type is
    # types.Function.
    if isinstance(fn, types.Function):
        msg = (
            "Returning function objects is not implemented. "
            f"getattr() was requested to return {fn} from attribute "
            f"'{lname}' of {obj}."
        )
        raise TypingError(msg)

    if fn is None:  # No attribute
        # if default is not _getattr_default then return the default
        if not (
            isinstance(default, types.NamedTuple)
            and default.instance_class == _getattr_default_type
        ):
            # it's not the marker default value, so return it
            sig = default(obj, name, default)

            def impl(cgctx, builder, sig, llargs):
                tmp = llargs[-1]
                cgctx.nrt.incref(builder, default, tmp)
                return tmp

        else:
            # else wire in raising an AttributeError
            fnty = tyctx.resolve_value_type(_getattr_raise_attr_exc)
            raise_sig = fnty.get_call_type(tyctx, (obj, name), {})
            sig = types.none(obj, name, default)

            def impl(cgctx, builder, sig, llargs):
                native_impl = cgctx.get_function(fnty, raise_sig)
                return native_impl(builder, llargs[:-1])

    else:  # Attribute present, wire in handing it back to the overload(getattr)
        sig = fn(obj, name, default)
        if isinstance(fn, types.BoundFunction):
            # It's a method on an object
            def impl(cgctx, builder, sig, ll_args):
                cast_type = fn.this
                casted = cgctx.cast(builder, ll_args[0], obj, cast_type)
                res = cgctx.get_bound_function(builder, casted, cast_type)
                cgctx.nrt.incref(builder, fn, res)
                return res

        else:
            # Else it's some other type of attribute.
            # Ensure typing calls occur at typing time, not at lowering
            attrty = tyctx.resolve_getattr(obj, lname)

            def impl(cgctx, builder, sig, ll_args):
                attr_impl = cgctx.get_getattr(obj, lname)
                res = attr_impl(cgctx, builder, obj, ll_args[0], lname)
                casted = cgctx.cast(builder, res, attrty, fn)
                cgctx.nrt.incref(builder, fn, casted)
                return casted

    return sig, impl


# These are marker objects to indicate "no default has been provided" in a call
_getattr_default_type = namedtuple("_getattr_default_type", "")
_getattr_default = _getattr_default_type()


# getattr with no default arg, obj is an open type and name is forced as a
# literal string. The _getattr_default marker is used to indicate "no default
# was provided".
@overload(getattr, prefer_literal=True)
def ol_getattr_2(obj, name):
    def impl(obj, name):
        return resolve_getattr(obj, name, _getattr_default)

    return impl


# getattr with default arg present, obj is an open type, name is forced as a
# literal string, the "default" is again an open type. Note that the CPython
# definition is: `getattr(object, name[, default]) -> value`, the `default`
# is not a kwarg.
@overload(getattr)
def ol_getattr_3(obj, name, default):
    def impl(obj, name, default):
        return resolve_getattr(obj, name, default)

    return impl


@overload(repr)
def ol_repr_generic(obj):
    missing_repr_format = f"<object type:{obj}>"

    def impl(obj):
        attr = "__repr__"
        if hasattr(obj, attr):
            return getattr(obj, attr)()
        else:
            # There's no __str__ or __repr__ defined for this object, return
            # something generic
            return missing_repr_format

    return impl


@overload(str)
def ol_str_generic(object=""):
    def impl(object=""):
        attr = "__str__"
        if hasattr(object, attr):
            return getattr(object, attr)()
        else:
            return repr(object)

    return impl



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
from cuda import core

# core.__version__ returns str like '1.0.1',
# '1.0.2.dev46+g16df11dab', etc.
CUDA_CORE_VERSION = tuple(int(i) for i in core.__version__.split(".") if i.isdigit())
CUDA_CORE_GT_0_6 = CUDA_CORE_VERSION >= (0, 6, 0)
CUDA_CORE_GE_1_0 = CUDA_CORE_VERSION >= (1, 0, 0)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from .cudadrv import devices, runtime
from numba_cuda_mlir.numba_cuda.core.codegen import Codegen, CodeLibrary
from numba_cuda_mlir.numba_cuda.cudadrv.linkable_code import LinkableCode
import os
import subprocess
import tempfile


def run_nvdisasm(cubin, flags):
    # nvdisasm only accepts input from a file, so we need to write out to a
    # temp file and clean up afterwards.
    fd = None
    fname = None
    try:
        fd, fname = tempfile.mkstemp()
        with open(fname, "wb") as f:
            f.write(cubin.code)

        try:
            cp = subprocess.run(
                ["nvdisasm", *flags, fname],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except FileNotFoundError as e:
            msg = (
                "nvdisasm has not been found. You may need "
                "to install the CUDA toolkit and ensure that "
                "it is available on your PATH.\n"
            )
            raise RuntimeError(msg) from e
        return cp.stdout.decode("utf-8")
    finally:
        if fd is not None:
            os.close(fd)
        if fname is not None:
            os.unlink(fname)


def disassemble_cubin(cubin):
    # Request lineinfo in disassembly
    flags = ["-gi"]
    return run_nvdisasm(cubin, flags)


def disassemble_cubin_for_cfg(cubin):
    # Request control flow graph in disassembly
    flags = ["-cfg"]
    return run_nvdisasm(cubin, flags)


class ExternalCodeLibrary(CodeLibrary):
    """Holds code produced externally, for linking with generated code."""

    def __init__(self, codegen, name):
        super().__init__(codegen, name)
        # Files to link
        self._linking_files = set()
        # Setup and teardown functions for the module.
        # The order is determined by the order they are added to the codelib.
        self._setup_functions = []
        self._teardown_functions = []

        self.use_cooperative = False

    @property
    def modules(self):
        # There are no LLVM IR modules in an ExternalCodeLibrary
        return set()

    def add_linking_file(self, path_or_obj):
        # Adding new files after finalization is prohibited, in case the list
        # of libraries has already been added to another code library; the
        # newly-added files would be omitted from their linking process.
        self._raise_if_finalized()

        if isinstance(path_or_obj, LinkableCode):
            if path_or_obj.setup_callback:
                self._setup_functions.append(path_or_obj.setup_callback)
            if path_or_obj.teardown_callback:
                self._teardown_functions.append(path_or_obj.teardown_callback)

        self._linking_files.add(path_or_obj)

    def add_ir_module(self, module):
        raise NotImplementedError("Cannot add LLVM IR to external code")

    def add_linking_library(self, library):
        raise NotImplementedError("Cannot add libraries to external code")

    def finalize(self):
        self._raise_if_finalized()
        self._finalized = True

    def get_asm_str(self):
        raise NotImplementedError("No assembly for external code")

    def get_llvm_str(self):
        raise NotImplementedError("No LLVM IR for external code")

    def get_function(self, name):
        raise NotImplementedError("Cannot get function from external code")


class JITCUDACodegen(Codegen):
    """
    Scaffolding codegen retained for the device-function declaration path
    (see numba_cuda_mlir.numba_cuda.compiler). It does not generate code -
    the MLIR pipeline produces code via numba_cuda_mlir.codegen instead.
    """

    def __init__(self, module_name):
        pass

    def _create_empty_module(self, name):
        raise NotImplementedError(
            "JITCUDACodegen does not generate LLVM IR modules; the MLIR "
            "pipeline produces code via numba_cuda_mlir.codegen instead."
        )

    def _add_module(self, module):
        pass

    def magic_tuple(self):
        """
        Return a tuple unambiguously describing the codegen behaviour.
        """
        ctx = devices.get_context()
        cc = ctx.device.compute_capability
        return (runtime.runtime.get_version(), cc)
